#!/usr/bin/env python
"""
Software Heritage functions.
"""

import logging
from urllib.parse import quote as urlquote

import requests


LOGGER = logging.getLogger(__name__)


def get_swhid_by_origin(origin):
    """
    Get a Software Heritage ID from an origin URL. This only includes the origin
    and repository ID.

    Args:
        origin:
            The origin URL of the software project.

    Returns:
        A Software Heritage ID string.
    """
    cache = get_swhid_by_origin.cache
    try:
        return cache[origin]
    except KeyError:
        pass

    timeout = 5
    url = f"https://archive.softwareheritage.org/api/1/origin/{origin}/get/"
    LOGGER.debug("Querying %s", url)
    resp = requests.get(url, timeout=timeout).json()
    try:
        resp = requests.get(resp["origin_visits_url"], timeout=timeout).json()
        resp = requests.get(resp[0]["snapshot_url"], timeout=timeout).json()
        resp = requests.get(
            resp["branches"]["HEAD"]["target_url"], timeout=timeout
        ).json()
        swhid = f'swh:1:dir:{resp["directory"]};origin={urlquote(origin)}'
        cache[origin] = swhid
        return swhid
    except KeyError:
        return None


get_swhid_by_origin.cache = {}


def get_swh_url_by_origin(origin):
    """
    Get a Software Heritage link for an origin URL.

    Args:
        origin:
            The origin URL of the software project.

    Returns:
        A Software Heritage URL. The page may not exist.
    """
    return f"https://archive.softwareheritage.org/browse/origin/?origin_url={urlquote(origin)}"


def swh_project_exists(origin):
    """
    Check if the project exists on Software Heritage.

    Args:
        origin:
            The origin URL of the software project.

    Returns:
        True if the project exists, False otherwise.
    """
    return get_swhid_by_origin(origin) is not None



#!/usr/bin/env python
"""
CodeMeta functions.
"""

import contextlib
import json
import logging

from ..exception import PrometaException
from ..file import update_content
from ..python.venv import VirtualEnvironment


LOGGER = logging.getLogger(__name__)


class CodeMetaError(PrometaException):
    """
    Custom exception raised when working with CodeMeta files.
    """


# https://codemeta.github.io/user-guide/
class CodeMeta:
    """
    Read and update CodeMeta files.
    """

    def __init__(self, project):
        """
        Args:
            project:
                The project object.
        """
        self.project = project
        self._data = None

    def load(self):
        """
        Load the data from the CodeMeta file.
        """
        path = self.project.codemeta_json_path
        LOGGER.debug("Loading CodeMeta data from %s", path)
        try:
            with path.open("rb") as handle:
                self._data = json.load(handle)
                return self._data
        except FileNotFoundError:
            return None
        except (OSError, json.JSONDecodeError) as err:
            raise CodeMetaError(err) from err

    def update_file(self):
        """
        Update the CodeMeta file with the currently loaded data. If the file
        already exists, a temporary file will be created and the merge tool will
        be used to update the existing file.
        """
        path = self.project.codemeta_json_path
        if self._data is None:
            LOGGER.warning("No new data to update %s", path)
            return
        content = json.dumps(self._data, indent=2, sort_keys=True)
        update_content(content, path)

    @property
    def data(self):
        """
        The CodeMeta data.
        """
        if self._data is None:
            self.load()
        return self._data

    @property
    def name(self):
        """
        The project name.
        """
        try:
            return self.data["name"]
        except TypeError:
            return None

    def _install_in_venv(self, venv):
        """
        Install CodeMetaPy in a Python virtual environment.

        Args:
            venv:
                An instance of VirtualEnvironment with its context currently
                open.
        """
        # TODO
        # Remove setuptools once CodeMetaPy is updated for Python > 3.12
        venv.run_pip_in_venv(
            [
                "install",
                "-U",
                "CodeMetaPy",
                "setuptools",
            ]
        )

    def _run_codemeta_in_venv(self, venv, args, **kwargs):
        """
        Run a codemeta command in the virtual environment.

        Args:
            venv:
                An instance of VirtualEnvironment with its context currently
                open.

            **kwargs:
                Keyword arguments passed through to run_python_in_venv().

        Returns:
            The return value of run_python_in_venv().
        """
        # TODO
        # Remove setuptools once CodeMetaPy is updated for Python > 3.12
        return venv.run_python_in_venv(["-m", "codemeta.codemeta", *args], **kwargs)

    def modify_codemeta_data(self, codemeta_data):
        """
        Modify the CodeMeta data. Override this method is a subclass to apply
        custom modifications.

        Args:
            codemeta_data:
                The input CodeMeta data.

        Returns:
            The possibly modified CodeMeta data.
        """
        return codemeta_data

    def update(self, version=None, cwd=None, venv=None):
        """
        Update the CodeMeta file.

        Args:
            version:
                The version to set. This is sometimes necessary to force a
                version due to SCM incrementing the version for unclean
                directories.

            cwd:
                The directory in which to run the codemeta command. If None, it
                will default to the parent directory of the target codemeta.json
                path.

            venv:
                An instance of VirtualEnvironment with the context currently
                open. If None, a new instance will be created.
        """
        codemeta_json_path = self.project.codemeta_json_path
        with contextlib.ExitStack() as stack:
            if venv is None:
                venv = stack.enter_context(
                    VirtualEnvironment(update_pip=True, inherit=False)
                )
            self._install_in_venv(venv)
            tmp_dir = venv.tmp_dir
            tmp_path = tmp_dir / codemeta_json_path.name
            if cwd is None:
                cwd = codemeta_json_path.parent

            self._run_codemeta_in_venv(venv, ["--enrich", "-O", str(tmp_path)], cwd=cwd)

            with tmp_path.open("rb") as handle:
                codemeta = json.load(handle)
            if version:
                codemeta["version"] = version
            codemeta = self.modify_codemeta_data(codemeta)
            codemeta_text = json.dumps(codemeta, indent=2, sort_keys=True)
            if not codemeta_text.endswith("\n"):
                codemeta_text += "\n"
            update_content(codemeta_text, codemeta_json_path)



#!/usr/bin/env python
"""\
Update project metadata.\
"""

import argparse
import logging
import pathlib
import subprocess
import sys

import yaml

from .common import NAME
from .config import CONFIG_FILE, CONFIG_EXAMPLE
from .exception import PrometaException
from .project import Project


LOGGER = logging.getLogger(__name__)


def main(args=None):
    """
    Main function.
    """
    parser = argparse.ArgumentParser(description=__doc__, prog=NAME)
    parser.add_argument("path", nargs="*", help="Path to project directory.")
    parser.add_argument(
        "--config",
        metavar="PATH",
        nargs="+",
        type=pathlib.Path,
        help=f"""
            By default, %(prog)s will search for configuration files named
            "{CONFIG_FILE}" or ".{CONFIG_FILE}" in the target directory and all
            of its parent directories, with precedence given to configuration
            files closest to the target directory. Additional configuration file
            paths can be passed with this option and they will take precedence
            over the detected configuration files. If multiple configuration
            paths are given with this command, their order determines their
            precedence.
            """,
    )
    parser.add_argument(
        "--gen-config",
        metavar="PATH",
        help=f"""
            Generate a configuration file template at the given path. If the
            path is "-", the file will be printed to STDOUT. Note that %(prog)s
            will only look for files named "{CONFIG_FILE}" or ".{CONFIG_FILE}".
            """,
    )
    parser.add_argument(
        "--list-config",
        choices=("all", "existing"),
        help="""
            List either all paths that will be scanned for configuration files
            for each given project, or only existing ones. The output is printed
            as a YAML file mapping project directory paths to lists of possible
            configuration files.
            """,
    )
    parser.add_argument(
        "--no-xdg",
        action="store_true",
        help="""
            Disable loading of configuration files in standard XDG locations.
            """,
    )
    parser.add_argument(
        "--trust",
        action="store_true",
        help="""
            It is possible to insert arbitrary command output into the README
            file. By default, %(prog)s will prompt the user for confirmation
            before running the command to prevent arbitrary code execution in
            the context of a collaborative environment. This option can be used
            to disable the prompt if the user trusts all of the commands in the
            README.
            """,
    )
    pargs = parser.parse_args(args=args)

    if pargs.gen_config:
        if pargs.gen_config == "-":
            print(CONFIG_EXAMPLE)
        else:
            path = pathlib.Path(pargs.gen_config).resolve()
            LOGGER.info("Creating %s", path)
            path.parent.mkdir(parents=True, exist_ok=True)
            path.write_text(CONFIG_EXAMPLE, encoding="utf-8")
        return

    config_args = {
        "custom_config_paths": pargs.config,
        "trust_commands": pargs.trust,
        "use_xdg": not pargs.no_xdg,
    }

    if pargs.list_config:
        paths = {}
        for path in pargs.path:
            proj = Project(path, **config_args)
            config_paths = (
                proj.config.possible_config_paths
                if pargs.list_config == "all"
                else proj.config.config_paths
            )
            paths[str(path)] = list(str(p) for p in config_paths)
        print(yaml.dump(paths))
        return

    for path in pargs.path:
        LOGGER.info("Updating %s", path)
        proj = Project(path, **config_args)
        proj.update()


def run_main(args=None):
    """
    Wrapper around main for exception handling.
    """
    logging.basicConfig(
        style="{", format="[{asctime}] {levelname} {message}", level=logging.INFO
    )
    try:
        main(args=args)
    except KeyboardInterrupt:
        pass
    except (PrometaException, subprocess.CalledProcessError) as err:
        sys.exit(err)


if __name__ == "__main__":
    run_main()



#!/usr/bin/env python
"""
Virtual environment functions.
"""

import contextlib
import logging
import pathlib
import sysconfig
import tempfile
import subprocess
import sys


LOGGER = logging.getLogger(__name__)


class VirtualEnvironment(contextlib.ExitStack):
    """
    Context manager for temporary Python virtual environments.
    """

    def __init__(self, update_pip=False, inherit=False):
        """
        Args:
            update_pip:
                If True, update pip in the virtual environment.

            inherit:
                If True, create a .pth file to inherit packages from the parent
                environment.
        """
        super().__init__()

        self.update_pip = update_pip
        self.inherit = inherit

        self.tmp_dir = None
        self.venv_exe = None
        self.venv_dir = None

    @property
    def sys_exe(self):
        """
        The system Python executable.
        """
        return pathlib.Path(sys.executable).resolve()

    def run_python_in_venv(self, args, **kwargs):
        """
        Run a Python command in the virtual environment.

        Args:
            args:
                The arguments to pass to the Python interpretter.

            **kwargs:
                Keyword arguments passed through to subprocess.run.

        Returns:
            The return value of subprocess.run.
        """
        cmd = [str(self.venv_exe), *args]
        kwargs.setdefault("check", True)
        LOGGER.debug("Running command: %s", cmd)
        return subprocess.run(cmd, **kwargs)  # pylint: disable=subprocess-run-check

    def run_pip_in_venv(self, args, **kwargs):
        """
        Run a pip command in the virtual environment. This is a wrapper around
        run_python_in_venv().

        Args:
            args:
                The arguments to pass to pip.

            **kwargs:
                Keyword arguments passed through to run_python_in_venv().

        Returns:
            The return value of run_python_in_venv().
        """
        return self.run_python_in_venv(["-m", "pip", *args], **kwargs)

    def __enter__(self):
        self.tmp_dir = pathlib.Path(self.enter_context(tempfile.TemporaryDirectory()))
        sys_exe = self.sys_exe
        self.venv_dir = self.tmp_dir / "venv"
        self.venv_exe = self.venv_dir / "bin" / sys_exe.name

        cmd = [str(sys_exe), "-m", "venv", str(self.venv_dir)]
        LOGGER.debug("Creating temporary virtual environment: %s", cmd)
        subprocess.run(cmd, check=True)

        if self.update_pip:
            LOGGER.debug("Updating pip in virtual environment: %s", cmd)
            self.run_pip_in_venv(["install", "-U", "pip"])

        if self.inherit:
            LOGGER.debug("Locating virtual environment's purelib directory.")
            child_purelib = (
                self.run_python_in_venv(
                    [
                        "-c",
                        'import sysconfig; print(sysconfig.get_paths()["purelib"])',
                    ],
                    stdout=subprocess.PIPE,
                )
                .stdout.decode()
                .strip()
            )
            parent_purelib = sysconfig.get_paths()["purelib"]
            pth_path = pathlib.Path(child_purelib) / "parent.pth"
            pth_path.write_text(parent_purelib, encoding="utf-8")

        return self



#!/usr/bin/env python
"""
All things specific to Python (pyproject.toml, PyPI, etc.)
"""

import difflib
import logging
import pathlib
import subprocess
import tomllib
from urllib.parse import quote as urlquote

import requests
import tomli_w
import tomli_w._writer
from spdx_license_list import LICENSES as SPDX_LICENSES
from trove_classifiers import classifiers as py_classifiers

from ..file import update_content
from .venv import VirtualEnvironment

LOGGER = logging.getLogger(__name__)


def get_pypi_url(name):
    """
    Get the URL to the project's page on PyPI if it exists.

    Args:
        name:
            The project name.

    Returns:
        The project URL, or None if it does not exist.
    """
    timeout = 5
    url = f"https://pypi.org/pypi/{urlquote(name)}/json"
    LOGGER.debug("Querying %s", url)
    resp = requests.get(url, timeout=timeout).json()
    try:
        return resp["info"]["package_url"]
    except KeyError:
        return None


def get_license_classifier(spdx_id):
    """
    Get the Python license classifier for the given SPDX license ID.

    Args:
        spdx_id:
            The SPXD license ID.

    Returns:
        The corresponding Python trove classifier if found, else None.
    """
    spdx_data = SPDX_LICENSES.get(spdx_id)
    if not spdx_data:
        LOGGER.warning("%s is not a recognized SPDX license ID.", spdx_id)
        return None
    name = spdx_data.name
    if spdx_data.osi_approved:
        expected = f"License :: OSI Approved :: {name}"
    else:
        expected = "License :: {name}"
    if expected in py_classifiers:
        return expected
    closest = difflib.get_close_matches(expected, py_classifiers)
    if not closest:
        LOGGER.warning(
            "Failed to map SPDX license ID %s to a Python trove classifier.", spdx_id
        )
        return None
    return closest[0]


def _join_names(given_names, family_names):
    """
    Join given and family names to a single string.

    Args:
        given_names:
            The given names.

        family_names:
            The family names.

    Returns:
        A string with the joined names.
    """
    if given_names and family_names:
        return f"{given_names} {family_names}"
    if given_names:
        return str(given_names)
    if family_names:
        return str(family_names)
    return None


def update_pyproject_toml(project):
    """
    Update the URLs in a pyproject.toml file.

    Args:
        project:
            The Project instance.
    """
    path = project.pyproject_toml_path
    data = tomllib.loads(path.read_text(encoding="utf-8"))

    urls = data["project"]["urls"]
    #  urls.clear()
    urls.update(project.urls)

    data["project"]["authors"] = [
        {
            "name": _join_names(author["given-names"], author["family-names"]),
            "email": author["email"],
        }
        for author in project.config.config["authors"]
    ]

    classifiers = set(
        classifier
        for classifier in data["project"].get("classifiers", [])
        if not classifier.startswith("License :: ")
    )

    spdx_id = project.spdx_license
    if spdx_id:
        license_classifier = get_license_classifier(spdx_id)
        if license_classifier:
            classifiers.add(license_classifier)
    else:
        LOGGER.warning("No license detected for project.")

    data["project"]["classifiers"] = sorted(classifiers)

    content = tomli_w.dumps(data, multiline_strings=True)
    update_content(content, path)
    return urls


def get_version(project_dir):
    """
    Get the version of a project from its directory.

    This will create a temporary virtual environment and install the project in
    it to get the version. This ensures that VCS-versions are correctly handled.

    This should not be necessary but at the time or writing the current version
    of CodeMetaPy fails to detect versions.
    """
    project_dir = pathlib.Path(project_dir).resolve()
    with VirtualEnvironment() as venv:
        venv.run_pip_in_venv(["install", "--no-deps", "-U", str(project_dir)])

        data = tomllib.loads(
            (project_dir / "pyproject.toml").read_text(encoding="utf-8")
        )
        name = data["project"]["name"]
        return (
            venv.run_python_in_venv(
                [
                    "-c",
                    f'from importlib.metadata import version; print(version("{name}"))',
                ],
                capture_output=True,
            )
            .stdout.decode()
            .strip()
        )



#!/usr/bin/env python3
"""
Common constants and functions.
"""

import logging


NAME = "prometa"
LOGGER = logging.getLogger(__name__)


def choose(items, include_none=False):
    """
    Prompt the user to choose an item from an iterable of items.

    Args:
        items:
            The iterable of items.

        include_none:
            If True, allow the user to choose None even if it is not in the
            list.

    Returns:
        The chosen item.
    """
    items = sorted(items)
    if include_none and None not in items:
        items.append(None)
    if not items:
        LOGGER.warning("No items to choose.")
        return None
    n_items = len(items)
    if n_items == 1:
        return items[0]
    while True:
        print("Choose one of the following:")
        for i, item in enumerate(items, start=1):
            print(f"{i:d} {item}")
        choice = input(f"Enter an integer in the range 1-{n_items:d} and press enter. ")
        try:
            choice = int(choice)
        except ValueError:
            LOGGER.error('"%s" is not a valid integer.', choice)
            continue
        if choice < 1 or choice > n_items:
            LOGGER.error("Invalid choice.")
            continue
        return items[choice - 1]



#!/usr/bin/env python
"""
Configuration.
"""


import logging
import os
import pathlib
import shlex

import gitlab
import yaml
from xdg.BaseDirectory import xdg_config_dirs

from .common import NAME
from .exception import PrometaException

LOGGER = logging.getLogger(__name__)
CONFIG_FILE = f"{NAME}.yaml"

CONFIG_EXAMPLE = """
# A list of authors. They will appear in various files (e.g. pyproject.toml,
# codemeta.json, CITATIONS.cff).
authors:
    # Given names (required)
  - given-names: John

    # Family names (required)
    family-names: Doe

    # Email (optional)
    email: john.doe@example.com

    # Affiliation (optional)
    affiliation: Example Institute

    # ORCID identifier (optional)
    orcid: XXXX-XXXX-XXXX-XXXX

    # HAL Open Science identifier (optional)
    hal: XXXXXXX

# If true, create missing CITATIONS.cff files.
citations_cff: true

# GitLab settings (optional)
gitlab:
  # Prometa uses python-gitlab to manage GitLab hooks that push code to other
  # open repositories (currently only Software Heritage). python-gitlab requires
  # both a configuration file and the name of the section in the configuration
  # file to use for a given project. For details, see the documentation:
  #
  # https://python-gitlab.readthedocs.io/en/stable/cli-usage.html#cli-configuration
  #
  # python-gitlab configuration file:
  config: path/to/python-gitlab.cfg

  # The section of the python-gitlab configuration file to use when retrieving
  # GitLab project data.
  section: somewhere

  # If true, use python-gitlab to update project hooks.
  update_hooks: false

  # Map GitLab hosts to their corresponding GitLab Pages URL formats. This map
  # will be used to generate documentation links when a "pages" job is detected
  # in the CI configuration file. The namespace and name parameters correspond
  # to those of the GitLab project.
  pages_urls:
    gitlab.com: "https://{namespace}.gitlab.io/{name}"

  # The regular expression for matching release tags. If given, a CI release job
  # will be created for tags that match this pattern. Omit this or set it to
  # null to disable release jobs.
  release_tag_regex: "^v."

  # Configure tags for GitLab CI jobs. This is a mapping of Python regular
  # expressions to lists of tags. Jobs that match the regular expressions will
  # be tagged with the corresponding tags. If multiple regular expressions match
  # a job then it will accumulate the tags.
  #
  # To apply the same tags to all jobs, use the regular expression ".".
  ci_tags:
      ".":
        - tag1
        - tag2
        - tag3

# By default, Prometa will attempt to detect each project's license using the
# spdx-matcher Python package. In some cases the detection fails (e.g. GPL v2
# and GPL v2-only use the same license text). This option can be set to an SPDX
# license identifier (https://spdx.org/licenses/) to force a particular license
# when the detection fails. If null or an empty strign then it will be ignored.
#
# Note that it will not download a new license file or modify the existing
# license file.
license: null

# The utility to use when merging changes. It must accept two file paths (the
# modified file and the original) and return non-zero exit status to indicate an
# error or abort.
merger: vimdiff


# The README interpolator can insert command output into the README. To prevent
# arbitrary command execution, Prometa will prompt the user to confirm a command
# before it is executed. This prompt can be surpressed for trusted READMEs by
# setting the following to true.
trust_commands: false
""".strip()


class ConfigError(PrometaException):
    """
    Custom error raised by the Config class.
    """


def _nested_values(data):
    """
    A generator over all nested values in a dict.
    """
    if not isinstance(data, dict):
        yield data
        return
    for value in data.values():
        if isinstance(value, dict):
            yield from _nested_values(value)
        else:
            yield value


def _nested_update(old_data, new_data, origin, new_path):
    """
    Updated nested values in a configuration dict recursively.

    Args:
        old_data:
            The dict to update in place.

        new_data:
            The data to add to old_dict.

        origin:
            A dict of the same layout as old_data but the values are updated
            with the filenames that provided them.

        new_path:
            The path from which new_data was loaded.
    """
    for key, new_value in new_data.items():
        old_value = old_data.get(key)
        if isinstance(new_value, dict) and isinstance(old_value, dict):
            _nested_update(old_value, new_value, origin[key], new_path)
        else:
            old_data[key] = new_value
            origin[key] = new_path


class Config:
    """
    Common non-derivable configuration.
    """

    def __init__(self, proj_path, custom_config_paths=None, use_xdg=True, **overrides):
        """
        Args:
            proj_path:
                The project path.

            custom_config_paths:
                An iterable over custon configuration file paths to use in
                addition to the standard configuration files that Prometa
                normally detects.

            use_xdg:
                If True, search for configuration files in standard XDG
                configuration directories. These files will be given the lowest
                priority.

            **overrides:
                Custom run-time overrides that take precedence over values in
                all discovered configuration files.
        """
        self.proj_path = pathlib.Path(proj_path).resolve()
        if not custom_config_paths:
            custom_config_paths = []
        self.custom_config_paths = [
            pathlib.Path(path).resolve() for path in custom_config_paths
        ]
        self._config = None
        self._origin = None
        self.use_xdg = use_xdg
        self.overrides = overrides

    @property
    def possible_config_paths(self):
        """
        A generator over the possible configuration file paths. Custom paths are
        yielded first, in the order they were given. Next the possible visible
        and hidden configuration paths in the current project directory and all
        of its parent directories are yielded, starting with the project
        directory and moving up to the root directory. Finally, if use_xdg is
        True, the standard XDG configuration directories are yielded, again in
        the standard order.

        The generator does not check if the paths exist but it will omit
        duplicate paths.
        """
        yielded = set()
        for path in self.custom_config_paths:
            if path not in yielded:
                yield path
                yielded.add(path)
        dir_path = self.proj_path
        while True:
            for fname in (CONFIG_FILE, f".{CONFIG_FILE}"):
                path = (dir_path / fname).resolve()
                if path not in yielded:
                    yield path
                    yielded.add(path)
            next_dir_path = dir_path.parent
            if dir_path != next_dir_path:
                dir_path = next_dir_path
            else:
                break
        if self.use_xdg:
            for config_dir in xdg_config_dirs:
                path = pathlib.Path(config_dir) / NAME / CONFIG_FILE
                if path not in yielded:
                    yield path
                    yielded.add(path)

    @property
    def config_paths(self):
        """
        A generator over existing configuration paths. It is a wrapper around
        possible_config_paths that checks for and logs existence.
        """
        for path in self.possible_config_paths:
            if path.exists():
                yield path

    @property
    def config(self):
        """
        The configuration file object. If None, there is no configuration file.

        Raises:
            ConfigError:
                One of the configuration files failed to load.
        """
        if self._config is None:
            configs = []
            for path in self.config_paths:
                LOGGER.info("Loading configuration file: %s", path)
                try:
                    with path.open("r", encoding="utf-8") as handle:
                        data = yaml.safe_load(handle)
                except (yaml.YAMLError, OSError) as err:
                    raise ConfigError(f"Failed to load {path}: {err}") from err
                configs.append((path, data))
            final_config = {}
            origin = {}
            for path, config in reversed(configs):
                _nested_update(final_config, config, origin, path)
            _nested_update(final_config, self.overrides, origin, "overrides")
            self._config = final_config
            self._origin = origin
        return self._config

    def get(self, *keys, default=None):
        """
        Retrieve a configuration file value. This will scan the loaded
        configuration files in order and return the first match.

        Args:
            *keys:
                The keys to the field. For example, to retrieve the value of
                "bar" under "foo", call get("foo", "bar"). Integers may also be
                used to index lists.

            default:
                The default value to return if no value was found.

        Returns:
            The target value, or the default if no value was found.
        """
        config = self.config
        origin = self._origin
        for i, key in enumerate(keys):
            try:
                config = config[key]
                try:
                    origin = origin[key]
                except TypeError:
                    pass
            except (KeyError, IndexError):
                return default
            except TypeError as err:
                paths = sorted(set(_nested_values(origin)))
                LOGGER.error(
                    "Failed to retrieve configuration value for %s: %s [%s]",
                    keys[: i + 1],
                    err,
                    ", ".join(shlex.quote(str(p)) for p in paths),
                )
                return default
        return config

    def _get_origin(self, *keys):
        """
        Similar to get() but returns the origin path for the given keys.
        """
        origin = self._origin
        if isinstance(origin, pathlib.Path):
            return origin
        for key in keys:
            origin = origin[key]
            if origin:
                return origin
        return None

    @property
    def gitlab(self):
        """
        The python-gitlab GitLab instance from the current configuration.
        """
        path = self.get("gitlab", "config")
        if path is None:
            LOGGER.warning("No python-gitlab configuration file specified.")
            path = os.getenv("PYTHON_GITLAB_CFG")
            if path is None:
                LOGGER.warning("PYTHON_GITLAB_CFG environment variable is unset.")
                return None
            path = pathlib.Path(path)
        else:
            origin_path = self._get_origin("gitlab", "config")
            if isinstance(origin_path, pathlib.Path):
                path = origin_path.resolve().parent.joinpath(path)
        path = path.resolve()
        section = self.get("gitlab", "section")
        if not section:
            LOGGER.warning(
                "No section specified for the python-gitlab configuration file: %s",
                path,
            )
        return gitlab.Gitlab.from_config(section, [str(path)])



#!/usr/bin/env python
"""
HAL open science functions.

https://api.archives-ouvertes.fr/docs/search
"""

import logging
from urllib.parse import quote as urlquote

import requests


LOGGER = logging.getLogger(__name__)


def get_hal_url_by_origin(origin):
    """
    Get the HAL ID by origin.
    """
    cache = get_hal_url_by_origin.cache
    try:
        return cache[origin]
    except KeyError:
        pass

    timeout = 5
    query = origin.split("://", 1)[1]
    search_url = f"https://api.archives-ouvertes.fr/search/?q={urlquote(query)}"
    LOGGER.debug("Querying %s", search_url)
    resp = requests.get(search_url, timeout=timeout).json()
    for doc in resp["response"]["docs"]:
        label = doc["label_s"]
        if f"origin={origin};" in label:
            url = doc["uri_s"]
            cache[origin] = url
            return url
    return None


get_hal_url_by_origin.cache = {}


def get_hal_id_from_url(url):
    """
    Extract the HAL ID from a HAL URL.
    """
    return url.rsplit("/", 1)[1]



#!/usr/bin/env python
"""
Git functions.
"""

import logging
import pathlib
import subprocess


LOGGER = logging.getLogger(__name__)


class GitlabRepo:
    """
    Basic functionality for retrieving Git information.
    """

    def __init__(self, path, remote="origin"):
        """
        Args:
            path:
                A path to the Git repository or any non-submodule path within
                it.

            remote:
                The remote repository name to use for generating URLs.
        """
        self.path = pathlib.Path(path).resolve()
        self.path = self.top_level
        self.remote = remote
        self._origin_url = None

    def run_cmd(self, cmd):
        """
        Run a git command and return its output.

        Args:
            cmd:
                The git sub-command and arguments.

        Returns:
            The command output.
        """
        cmd = ("git", "-C", str(self.path), *cmd)
        LOGGER.debug("Running command: %s", cmd)
        return (
            subprocess.run(cmd, check=True, capture_output=True).stdout.decode().strip()
        )

    @property
    def top_level(self):
        """
        The top-level directory.
        """
        path = self.run_cmd(("rev-parse", "--show-toplevel"))
        return pathlib.Path(path).resolve()

    @property
    def remote_url(self):
        """
        The origin URL.
        """
        if self._origin_url is None:
            self._origin_url = self.run_cmd(("config", "--get", "remote.origin.url"))
        return self._origin_url

    @property
    def parsed_origin(self):
        """
        A 3-tuple of the origin's host, namespace and project name.
        """
        host, subpath = self.remote_url.split(":", 1)
        subpath = subpath.rsplit(".", 1)[0]
        namespace, name = subpath.split("/", 1)
        host = host.split("@", 1)[1]
        return host, namespace, name

    @property
    def public_url(self):
        """
        The URL for the main project page.
        """
        host, namespace, name = self.parsed_origin
        return f"https://{host}/{namespace}/{name}"

    def get_section_url(self, section):
        """
        The URL to one of the various GitLab sections, e.g. "blob/main" or
        "packages".
        """
        section = section.lstrip()
        return f"{self.public_url}/-/{section}"

    @property
    def public_git_url(self):
        """
        The publically accessible HTTPS URL.
        """
        return f"{self.public_url}.git"

    def get_main_blob_url(self, path):
        """
        Get the main branch URL to the given path.
        """
        path = path.lstrip("/")
        return self.get_section_url(f"blob/main/{path}")

    @property
    def readme_url(self):
        """
        The URL to the README.
        """
        return self.get_main_blob_url("README.md")

    def get_gitlab_project(self, glab):
        """
        Get the python-gitlab Project instance from a python-gitlab GitLab
        instance.
        """
        _host, namespace, name = self.parsed_origin
        return glab.projects.get(f"{namespace}/{name}")



#!/usr/bin/env python
"""
CodeMeta functions.
"""

import contextlib
import logging
import pathlib
import shutil
import tempfile
import tomllib

import tomli_w
import tomli_w._writer

from .codemeta import CodeMeta
from ..id.orcid import get_orcid_url
from ..python.venv import VirtualEnvironment


LOGGER = logging.getLogger(__name__)


class PyprojectCodeMeta(CodeMeta):
    """
    Update CodeMeta files from pyproject.toml files.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._name = None

    # TODO
    # Remove once CodeMetaPy is updated.
    @contextlib.contextmanager
    def _modified_pyproject_toml(self):
        """
        Context manager to create a temporary modified pyproject.toml file that
        works around current bugs in CodeMetaPy. The origin of the problem is
        that the project metadata now returns custom objects for the "readme"
        and "license" fields while CodeMetaPy still expects these to be strings.

        Returns:
            The temporary path to the modified pyproject.toml file.
        """
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_dir = pathlib.Path(tmp_dir)
            git_dir = self.project.git_repo.path
            shutil.copytree(git_dir, tmp_dir, dirs_exist_ok=True, symlinks=True)
            ppt_path = git_dir / "pyproject.toml"
            tmp_ppt_path = tmp_dir / ppt_path.name

            ppt_data = tomllib.loads(ppt_path.read_text(encoding="utf-8"))
            for field in ("license", "readme"):
                try:
                    del ppt_data["project"][field]
                except KeyError:
                    pass
            content = tomli_w.dumps(ppt_data, multiline_strings=True)
            tmp_ppt_path.write_text(content, encoding="utf-8")
            self._name = ppt_data["project"]["name"]

            yield tmp_ppt_path

    def modify_codemeta_data(self, codemeta_data):
        author_data = []
        for author in self.project.config.get("authors", default=[]):
            data = {
                "@type": "Person",
                "email": author["email"],
                "givenName": author["given-names"],
                "familyName": author["family-names"],
            }
            orcid = author.get("orcid")
            if orcid:
                data["@id"] = get_orcid_url(orcid)
            author_data.append(data)
        codemeta_data["author"] = author_data

        codemeta_data["name"] = self._name
        codemeta_data["maintainer"] = author_data[0]
        codemeta_data["readme"] = self.project.git_repo.readme_url

        for cont in codemeta_data.get("contributor", []):
            if all(
                cont[key] == author_data[0][key] for key in ("familyName", "givenName")
            ):
                cont.clear()
                cont.update(author_data[0])

        return codemeta_data

    def update(self, version=None, cwd=None, venv=None):
        with contextlib.ExitStack() as stack:
            if venv is None:
                venv = stack.enter_context(
                    VirtualEnvironment(update_pip=True, inherit=False)
                )
            tmp_ppt_path = stack.enter_context(self._modified_pyproject_toml())
            # Install the package to ensure that all metadata is detected (e.g.
            # script descriptions).
            venv.run_pip_in_venv(["install", "-U", "-e", str(tmp_ppt_path.parent)])
            super().update(version=version, cwd=tmp_ppt_path.parent, venv=venv)



#!/usr/bin/env python3
"""
File operations.
"""

import filecmp
import logging
import pathlib
import shutil
import subprocess
import tempfile


LOGGER = logging.getLogger(__name__)


def diff(path_1, path_2, differ="vimdiff"):
    """
    Diff 2 paths.
    """
    if not path_2.exists():
        LOGGER.debug("Copying %s to %s", path_1, path_2)
        shutil.copy(path_1, path_2)
        return
    if not filecmp.cmp(path_1, path_2, shallow=False):
        #  if path_1.read_bytes().rstrip(b'\n') == path_2.read_bytes().rstrip(b'\n'):
        #      return
        cmd = [differ, str(path_1), str(path_2)]
        LOGGER.info("Diffing %s and %s", path_1, path_2)
        subprocess.run(cmd, check=True)


def update_content(content, path, encoding="utf-8", **kwargs):
    """
    Interactively compare and merge new content. If the target path does not
    exist, the content will be written directly to it.

    Args:
        content:
            The new content to merge.

        path:
            The target path.

        encoding:
            The file encoding to use when writing the content. If None, the
            content is assumed to be bytes.

        **kwargs:
            Keyword arguments passed through to diff().
    """
    path = pathlib.Path(path).resolve()
    if not path.exists():
        LOGGER.debug("Writing content to %s", path)
        path.parent.mkdir(parents=True, exist_ok=True)
        if encoding is None:
            path.write_bytes(content)
        else:
            path.write_text(content, encoding=encoding)
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = pathlib.Path(tmp_dir)
        tmp_path = tmp_dir / path.with_stem("new").name
        LOGGER.debug("Writing content to %s", tmp_path)
        if encoding is None:
            tmp_path.write_bytes(content)
        else:
            tmp_path.write_text(content, encoding=encoding)
        diff(tmp_path, path, **kwargs)



#!/usr/bin/env python
"""
Project class.
"""

import logging

import spdx_matcher
from spdx_license_list import LICENSES as SPDX_LICENSES

from .citation import Citation
from .common import choose
from .codemeta.codemeta import CodeMeta
from .codemeta.pyproject import PyprojectCodeMeta
from .config import Config
from .gitlab.ci import GitlabCI
from .gitlab.repo import GitlabRepo
from .python.common import update_pyproject_toml, get_version
from .readme import ReadmeInserter


LOGGER = logging.getLogger(__name__)


class Project:
    """
    Project class.
    """

    def __init__(self, path, **config_kwargs):
        """
        Args:
            path:
                The path to the project.

            **config_kwargs:
                Keyword arguments passed through to Config.
        """
        self.git_repo = GitlabRepo(path)
        self.gitlab_ci = GitlabCI(self)
        self.config = Config(self.git_repo.path, **config_kwargs)
        self.codemeta = CodeMeta(self)

    @property
    def readme_md_path(self):
        """
        The README.md path.
        """
        return self.git_repo.path / "README.md"

    @property
    def pyproject_toml_path(self):
        """
        The pyproject.toml path.
        """
        return self.git_repo.path / "pyproject.toml"

    @property
    def license_txt_path(self):
        """
        The path to the LICENSE.txt file.
        """
        return self.git_repo.path / "LICENSE.txt"

    @property
    def spdx_license(self):
        """
        The detected SPDX license name.
        """
        config_license = self.config.get("license")
        if config_license:
            LOGGER.info("Using license from configuration file: %s", config_license)
            if config_license not in SPDX_LICENSES:
                LOGGER.warning('"%s" is not a recognized SPDX license', config_license)
            return config_license
        path = self.license_txt_path
        LOGGER.info("Attempting to detect license from %s", path)
        try:
            detected, percent = spdx_matcher.analyse_license_text(
                path.read_text(encoding="utf-8")
            )
        except FileNotFoundError:
            LOGGER.error("No license file found at %s", path)
            return None
        licenses = list(detected["licenses"])
        if not licenses:
            LOGGER.error("Failed to detect license in %s", path)
            return None
        lic = choose(licenses)
        if percent < 0.9:
            LOGGER.warning(
                "Detected %s license in %s but certainty is only %(0.0f) %",
                lic,
                path,
                percent * 100.0,
            )
        return lic

    @property
    def codemeta_json_path(self):
        """
        The path to the codemeta.json file.
        """
        return self.git_repo.path / "codemeta.json"

    @property
    def citation_cff_path(self):
        """
        The path to the CITATION.cff file.
        """
        return self.git_repo.path / "CITATION.cff"

    @property
    def urls(self):
        """
        A dict of URLs for the project.
        """
        host, namespace, name = self.git_repo.parsed_origin
        homepage = f"https://{host}/{namespace}/{name}"

        urls = {}
        urls["Homepage"] = homepage
        urls["Source"] = f"{homepage}.git"
        if self.gitlab_ci.data.get("pages"):
            pages_fmt = self.config.get("gitlab", "pages_urls", host)
            if pages_fmt:
                urls["Documentation"] = pages_fmt.format(namespace=namespace, name=name)
        urls["Issues"] = f"{homepage}/-/issues"
        return urls

    def update(self):
        """
        Update project metadata.
        """
        if self.pyproject_toml_path.exists():
            # TODO
            # Find a better way to automate version management while updating
            # the files. The issue is that prometa has to modify several files
            # but these modifications are visible to the VCS. This causes tools
            # such as setuptools-scm to miscalculate the version because it
            # doesn't know that the changes will be merged into the previous
            # commit.
            version = get_version(self.git_repo.path)
            update_pyproject_toml(self)
            PyprojectCodeMeta(self).update(version=version)

        if self.codemeta_json_path.exists():
            Citation(self).update()

        readme_inserter = ReadmeInserter(self)
        readme_inserter.update(self.readme_md_path)

        self.gitlab_ci.update()

    @property
    def name(self):
        """
        The project name. It will use the value in codemeta.name if it exists,
        otherwise it will use the name of the project's directory.
        """
        name = self.codemeta.name
        if name is not None:
            return name
        return self.git_repo.path.name



#!/usr/bin/env python
"""
Update the CI file.
"""

import logging
import re

import yaml

from ..file import update_content

LOGGER = logging.getLogger(__name__)


class GitlabCI:
    """
    Wrapper around the gitlab-ci file.
    """

    def __init__(self, project, path=".gitlab-ci.yml"):
        """
        Args:
            project:
                A Project instance.

            path:
                The path to the gitlab-ci file, relative to the repository root
                directory.
        """
        self.project = project
        self.path = project.git_repo.path / path
        self._data = None

    @property
    def data(self):
        """
        The CI configuration data.
        """
        if self._data is None:
            self._data = self.load()
        return self._data

    def load(self):
        """
        Load the file data.

        Returns:
            The loaded data, or an empty dict if the file does not exist.
        """
        try:
            with self.path.open("rb") as handle:
                return yaml.safe_load(handle)
        except FileNotFoundError:
            return {}

    def update_pages(self):
        """
        Update the pages job.
        """
        pages = self.data.get("pages")
        if not pages:
            return
        pages["artifacts"] = {"paths": ["public"]}
        pages["only"] = ["main"]
        pages["stage"] = "deploy"

    def add_register_pip_pkg(self):
        """
        Add a jot to register a pip package.
        """
        key = "register_pip_pkg"
        if self.project.pyproject_toml_path.exists():
            self.data[key] = {
                "image": "python:latest",
                "only": ["main"],
                "script": [
                    "pip install build twine",
                    "python -m build",
                    "TWINE_PASSWORD=${CI_JOB_TOKEN} TWINE_USERNAME=gitlab-ci-token "
                    "python -m twine upload --repository-url "
                    "${CI_API_V4_URL}/projects/${CI_PROJECT_ID}/packages/pypi "
                    "dist/*",
                ],
                "stage": "deploy",
            }
        else:
            self.data.pop(key)

    def add_release_job(self):
        """
        Add a release job that triggers when Git tags are pushed.
        """
        key = "release_job"
        regex = self.project.config.get("gitlab", "release_tag_regex")
        if regex:
            self.data[key] = {
                "image": "registry.gitlab.com/gitlab-org/release-cli:latest",
                "release": {
                    "description": "Release $CI_COMMIT_TAG",
                    "tag_name": "$CI_COMMIT_TAG",
                },
                "rules": [{"if": f"$CI_COMMIT_TAG =~ /{regex}/"}],
                "script": ['echo "running release_job"'],
                "stage": "release",
            }
        else:
            self.data.pop(key, None)

    def _manage_swh_hook(self, gproj):
        """
        Manage the SWH webhooks.

        Args:
            gproj:
                The python-gitlab Project instance.
        """
        # Reset to releases_events when SoftwareHeritage updates their API.
        #  events_key = 'releases_events'
        events_key = "tag_push_events"
        url = "https://archive.softwareheritage.org/api/1/origin/save/webhook/gitlab/"

        swh_hook = None
        add = self.project.codemeta_json_path.exists()

        for hook in gproj.hooks.list(iterator=True):
            if hook.attributes["url"] == url:
                if add and swh_hook is None:
                    swh_hook = hook
                else:
                    LOGGER.info("Deleting SWH hook for %s.", gproj.name)
                    hook.delete()

        if add and swh_hook is None:
            LOGGER.info("Creating SWH webhook for %s", gproj.name)
            swh_hook = gproj.hooks.create({"url": url, events_key: True})

        if swh_hook:
            changed = False
            for key, value in swh_hook.attributes.items():
                if key.endswith("_events"):
                    expected_value = key == events_key
                    if value != expected_value:
                        changed = True
                        setattr(swh_hook, key, expected_value)
            if changed:
                LOGGER.info("Updating SWH hook for %s.", gproj.name)
                swh_hook.save()

    def manage_hooks(self):
        """
        Add or remove hooks depending on the current repository configuration.
        """
        if not self.project.config.get("gitlab", "manage_hooks", default=False):
            return
        with self.project.config.gitlab as glab:
            glab.auth()
            gproj = self.project.git_repo.get_gitlab_project(glab)
            self._manage_swh_hook(gproj)

    def add_stages(self):
        """
        Add the list of stages for all added jobs.
        """
        recognized_stages = [".pre", "build", "test", "release", "deploy", ".post"]
        stages = set()
        for key, value in self.data.items():
            if not isinstance(value, dict):
                continue
            stage = value.get("stage")
            if stage is None:
                LOGGER.error("No stage specified for job %s in %s", key, self.path)
                continue
            stages.add(stage)
        self.data["stages"] = [stage for stage in recognized_stages if stage in stages]

    def add_tags(self):
        """
        Add runner tags. This will deduplicate tags and also ensure that jobs
        using the same tags reference each other in the YAML output.
        """
        tag_map = self.project.config.get("gitlab", "ci_tags")
        if tag_map:
            tag_map = tuple(
                (re.compile(regex), tags) for regex, tags in tag_map.items()
            )
            previous_sets = []
            for name, data in self.data.items():
                if not isinstance(data, dict):
                    continue
                collected_tags = set()
                for regex, new_tags in tag_map:
                    if regex.search(name):
                        collected_tags.update(new_tags)
                collected_tags = sorted(collected_tags)
                # PyYAML seems to only emit references when nested items refer
                # to the same instance so ensure that identical lists do.
                for prev in previous_sets:
                    if collected_tags == prev:
                        collected_tags = prev
                        break
                else:
                    previous_sets.append(collected_tags)
                self.data[name]["tags"] = collected_tags

    def update(self):
        """
        Update the CI file by detecting which jobs should be added.
        """
        self.update_pages()
        self.add_register_pip_pkg()
        self.add_release_job()
        self.add_stages()
        self.add_tags()
        update_content(yaml.dump(self.data), self.path)
        self.manage_hooks()
